#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "matrix.h"

#define EPSILON 1e-6
#define abs(a) (((a) > 0) ? (a) : (-a))

/* 
 * Matrix Library Code.  Interface isn't great but it's good
 * enough for non-OO applications. 
 */

int mSolve (double *x, double *A, double *b, int n) {
/*
 * int mSolve (double *x, double *A, double *b, int n);
 *
 * Solves a system of n equations Ax = b where A is a square matrix
 * containing variable coefficients and b is the right side of the
 * system.
 */
 
  double *P, *L, *U;
  double *c, *d;
  int i, j, k;
  double s;

  /* allocate memory for all our matrices. */ 
  P = (double *) malloc (n * n * sizeof (double));
  L = (double *) malloc (n * n * sizeof (double));
  U = (double *) malloc (n * n * sizeof (double));
  d = (double *) malloc (n * sizeof (double));
  c = (double *) malloc (n * sizeof (double));

  if (!P || !L || !U || !d || !c) {
    fprintf (stderr, "mSolve: could not allocate temp matrices\n");
    return (-1);
  }
  if (mLUdecomp (A, L, U, P, n) == -1) {
    free (P);
    free (L);
    free (U);
    free (d);
    free (c);
    return (-1);
  }
  
  mMul (d, P, b, n, n, 1);    /* d = P*b */
  
 /* forward elimination on L */
  mZero (c, n, 1); 
  for (k = 0; k<n; k++) {
    s = 0;
    for (j = 0; j<=k-1; j++) {
      s += L[k*n+j] * c[j]; 
    }
    c[k] = d[k] - s;
  }
  

 /* back substition with U */

  mZero (x, n, 1);
  for (k = n-1; k >= 0; k--) {
    s = 0;
    for (j = k+1; j<n; j++) 
      s += U[k*n+j] * x[j];
    x[k] = (c[k] - s) / U[k*n +k];  
  }
  
  free (P);
  free (L);
  free (U);
  free (d);
  free (c);
  return 0;
}

void mTranspose (double *mout, double *min, int m, int n) {
/*
 * void mTranspose (double *mout, double *min, int m, int n);
 *
 * Transposes the m x n matrix min into mout. 
 */
  
  int i, j;

  for (i=0; i<m; i++) 
    for (j=0; j<n; j++) 
       mout[m*j+i] = min[n*i+j];  

}

void mStar (double *M, double *v) {
/*
 *  Creates a matrix that looks like this:
 *
 *  0  -v2  v1   0
 *  v2   0 -v0   0
 * -v1  v0   0   0
 *  0    0   0   1
 *
 * from a vector.
 */

  mZero(M,4,4);
  M[4*0 + 1] = -v[2];
  M[4*0 + 2] =  v[1];

  M[4*1 + 0] =  v[2];
  M[4*1 + 2] = -v[0];

  M[4*2 + 0] = -v[1];
  M[4*2 + 1] =  v[0];

}


void mMul (double *out, double *m1, double *m2, int n, int m, int p) {
/*
 * void mMul (double *out, double *m1, double *m2, int n, int m, int p);
 *
 * General matrix multiplication routine.  Multiplies n x m matrix m1 by
 * m x p m2 and stores the n x p result in out.
 */

  int i, j, k;
  double d;

  for (i=0; i<p; i++) {           
    for (j=0; j<n; j++) {         
      d = 0.0;
      for (k=0; k<m; k++) 
        d += m2[p*k + i] * m1[n*j  + k];
      out[p*j + i] = d; 
    }
  }
}
 
int mLUdecomp (double *M, double *L, double *U, double *P, int n) {
/*
 * int mLUdecomp (double *M, double *L, double *U, double *P, int n);
 *
 * Routine for factoring an NxN matrix M into L and U (diagonal matrices)
 * and a permutation matrix P.
 *
 */

  int i, j, k;
  double m;
  int pivot, pind; 

  double *A;

 /* Make a copy of M so we can muck around with it */  

  A = (double *) malloc (sizeof(double) * n * n);
  if (!A) {
    fprintf (stderr, "mLUdecomp: cannot malloc temporary matrix\n");
    return -1;
  }
  memcpy (A, M, n*n*sizeof(double));  

 /* initialize P and LU */
  mIdent (P, n);
  mIdent (L, n);
  mZero  (U, n, n);  

  for (k=0; k<n; k++) {
    for (i=k, pivot=0, m=0.0; i<n; i++) {    /* look for largest pivot */    
      if (abs(A[n*i+k]) >= EPSILON) {      
        pivot = 1;
        if (abs(m) < abs(A[n*i+k])) { 
          pind = i;
          m = A[n*i+k];      
        }
      }
    }
    if (!pivot) {                          /* M's not invertible. */
      fprintf (stderr, "mLUdecomp: M is singular in column %d\n", k);
      free (A);
      return;
    }
    /* pivot found, exchange row k with row pind */ 
    mExchangeRow (A, k, pind, n);
    mExchangeRow (P, k, pind, n);
    mExchangeRow (L, k, pind, k);

    for (i=k+1; i<n; i++) {
      L[i*n+k] = A[n*i + k] / A[n*k + k];
      for (j=k+1; j<n; j++) 
        A[n*i+j] = A[n*i+j] - L[n*i+k] * A[n*k+j];
    }
    for (i=k; i<n; i++) {
      U[k*n+i] = (abs(A[n*k+i]) >= EPSILON) ? A[n*k + i] : 0; 
    } 

  }
  free (A);
  return 0;
}


void mZero (double *M, int m, int n) {

  int i,j;

  for (i=0; i<m; i++) 
    for (j=0; j<n; j++) 
       M[n*i + j] = 0.0;
}


void mIdent (double *M, int n) {

  int i, j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      if (i == j) 
        M[n*i + j] = 1.0;
      else 
        M[n*i + j] = 0.0;
    }
  }
}

void mExchangeRow (double *M, int m1, int m2, int n) {

  int i;
  double temp;
 
  for (i=0; i<n; i++) {
    temp = M[m1*n + i];
    M[m1*n + i] = M[m2*n + i];
    M[m2*n + i] = temp;
  } 
}

void mCreateTranslate (double *M, double X, double Y, double Z) {
/*
 * Creates a 4x4 translation matrix.
 * 
 */

  mIdent(M, 4);

  M[4*0+3]=X;
  M[4*1+3]=Y;
  M[4*1+3]=Z;
}

void mCreateRotate (double *M, double X, double Y, double Z) {
/*
 * should be around *some* axis :)
 * 
 */

 mIdent(M,4);
}

void mPrintMatrix (double *M, int m, int n, char *label) {
/*
 *  Prints a matrix, formatted similar to MATLAB's 
 *  output.
 */

  int i,j;

  double s;
  printf ("%s = \n\n", label);

  for (j=0; j<m; j++) { 
    printf ("  ");
    for (i=0; i<n; i++) {
      s = M[j*n+i];                
      if (s == -0.0) s = 0.0;    /* doubles are strange.          */
      printf ("%10.6f ",s);
    }
    printf ("\n");
  } 
  printf ("\n");
}
