/*
 * Simulation routines.  Mostly calls to ode solver.
 *
 * This code is based on "An Introduction to Physically Based
 * Modeling: Rigid Body Simulation I - Unconstrained Rigid Body 
 * Dynamics" by David Baraff.
 *
 */

#include "ode.h"
#include "polygon.h"
#include "matrix.h"
#include "objmgr.h"
#include "sim.h"
#include "force.h"
#include "icollision.h"

int num_objs;
extern OBJ_NODE ObjList;

void InitSim (void) {
/* void InitSim (void);
 * 
 * Does whatever setup code needed.  Going to use it as a cheesy 
 * constructor type thing so I don't have to use global variables.
 */ 
  num_objs = SizeList();
}

void State_to_Array (OBJECT *o, double *y) {

  int i,j;

 /* copy position vector */
  *y++ = o->x[0]; *y++ = o->x[1]; *y++ = o->x[2];

 /* copy rotation matrix */
  for (i = 0; i<3; i++) 
    for (j=0; j<3; j++) 
      *y++ = o->R[i*4+j];

  /* copy momentums */
  *y++ = o->P[0]; *y++ = o->P[1]; *y++ = o->P[2];
  *y++ = o->L[0]; *y++ = o->L[1]; *y++ = o->L[2];
}

void Array_to_State (OBJECT *o, double *y) {

  int i,j;
  double M[16], R[16];

  o->x[0] = *y++; o->x[1] = *y++; o->x[2] = *y++; 
  for (i = 0; i<3; i++) 
    for (j=0; j<3; j++) 
      o->R[i*4+j] = *y++;

  o->P[0] = *y++; o->P[1] = *y++; o->P[2] = *y++; 
  o->L[0] = *y++; o->L[1] = *y++; o->L[2] = *y++; 

 /* v(t) = P(t) / M */
  for (i=0; i<4; i++)
    o->v[i] = o->P[i] * o->mass_inv;

 /* I^-1 (t) = R(t) Ibody^-1 R(t)^T */
  mTranspose (R,o->R, 4, 4);
  mMul (M, o->ibody_inv, R, 4, 4, 4); 
  mMul (o->i_inv, o->R, M, 4, 4, 4);

 /* w(t) = I^-1(t) L(t) */
  mMul (o->omega, o->i_inv, o->L, 4, 4, 1);

}

void Array_to_Bodies (double *y) {
  int i;

  for (i=0; i<num_objs; i++) 
    Array_to_State(IndexList(i), &y[i*STATE_SIZE]);
}

void Bodies_to_Array (double *y) {
  int i;

  for (i = 0; i<num_objs; i++)
    State_to_Array(IndexList(i), &y[i*STATE_SIZE]);
}

void dydt (double t, double *y, double *ydot) {

  int i,j,k; OBJECT *o; int colfound;

  Array_to_Bodies(y);
  
  for (k=0; k<num_objs; k++) {
    o = IndexList(k);    
    vec4init (o->force);
    vec4init (o->torque);
    for (i=0; i<o->num_vertices; i++) {
      mat4vec4mul (&(o->verts_world[i*4]), o->R, &(o->vertices[i*4]));
      for (j=0; j<3; j++) {
        o->verts_world[i*4+j] = o->verts_world[i*4+j] + o->x[j];
      }
    }
  }

  update_translations( ObjList.next );
  colfound = compute_collisions( ObjList.next );
  if (colfound) Bodies_to_Array(y);
  for (i=0; i<num_objs; i++) {
    if (!colfound) Calc_Forces(t, IndexList(i));
    ddt_State_to_Array(IndexList(i), &ydot[i * STATE_SIZE]);
  }
}


void ddt_State_to_Array(OBJECT *o, double *ydot) {

  int i,j;
  double SW[16];
  double Rdot[16];
 
  /* x'(t) */ 
  *ydot++ = o->v[0]; 
  *ydot++ = o->v[1]; 
  *ydot++ = o->v[2]; 

  mStar(SW, o->omega);
 
  mMul (Rdot, SW, o->R, 4, 4, 4);

  /* R'(t) */  
  for (i = 0; i<3; i++)
    for (j =0; j<3; j++) 
      *ydot++ = Rdot[i*4+j];

  /* P'(t) */
  *ydot++ = o->force[0];
  *ydot++ = o->force[1];
  *ydot++ = o->force[2];

  /* L'(t) */
  *ydot++ = o->torque[0];
  *ydot++ = o->torque[1];
  *ydot++ = o->torque[2];
  
}
